#! /usr/bin/python3 -u

import os
import subprocess
import time
import sys
from fnmatch import fnmatch
from optparse import OptionParser


# reads required inputs and executable file from xrsl
def parse_xrsl(xrsl):
    with open(xrsl) as f:
        lines = f.readlines()

    if 'executable' in lines[1]:
        executable = lines[1].strip().split('=')[1].strip()[1:-2]
    else:
        print('Cannot find executable')

    inputs = []
    if 'inputfiles' in lines[2].lower():
        i = 3
        while lines[i].strip() != ')':
            inputs.append(eval(lines[i].strip().replace(' ', ',')))
            i += 1
    else:
        print('Cannot find inputFiles')

    return {'exec': executable, 'inputs': inputs}


# copies required files to 'results' folder, executes the job, cleans the inputs
def exec_xrsl(xrsl, cleanup=True):
    parsed = parse_xrsl(xrsl)

    if os.path.isdir('results'):
        subprocess.call('mv results results.{}'.format(time.time()), shell=True)
    subprocess.call('mkdir results', shell=True)

    for f in parsed['inputs']:
        if '/' in f[0]:
            folder = f[0][:f[0].rfind('/')]
            subprocess.call('mkdir -p results/{}'.format(folder), shell=True)
        if f[1]:
            subprocess.call('cp {} results/{}'.format(f[1], f[0]), shell=True)
        else:
            subprocess.call('cp {} results/{}'.format(f[0], f[0]), shell=True)

    subprocess.call('chmod +x results/{}'.format(parsed['exec']), shell=True)
    root = os.getcwd()
    os.chdir('results')

    start = time.time()
    subprocess.call('bash {}'.format(parsed['exec']), shell=True)
    finish = time.time()

    with open('time.txt', 'w') as f:
        print(finish-start, file=f)

    if cleanup:
        for f in parsed['inputs']:
            subprocess.call('rm {}'.format(f[0]), shell=True)
    os.chdir(root)


def find_jobs(xrsl, path):
    """Finds xrsl files in the current folder and its subfolders."""
    jobs = []

    for subdir, dirs, files in os.walk(path):
        if xrsl in files:
            jobs.append({
                'file': xrsl,
                'folder': subdir
            })

    return jobs


def is_finished(job):
    """Checks if there is a result folder with an .out file with size > 0."""
    path = job['folder'] + '/results'
    if os.path.isdir(path):
        for f in os.listdir(path):
            if fnmatch(f, '*.out'):
                return os.stat(path+'/'+f).st_size > 42
            elif fnmatch(f, '*.rep'):
                return os.stat(path+'/'+f).st_size > 42
    else:
        return False


if __name__ == '__main__':

    parser = OptionParser()

    parser.add_option('-X', '--xrsl', dest='xrslfile', help='name of xrsl files to execute', metavar='FILE')

    parser.add_option('--redo', dest='redo', default=False, action='store_true',
                      help='repeat the jobs that already have a result')

    parser.add_option('--no_cleanup', dest='cleanup', default=True, action='store_false',
                      help='do not delete the links and temporary files in the results folder')

    parser.add_option('--no_confirm', dest='confirm', default=True, action='store_false',
                      help='do not require confirmation before starting')

    options, argv = parser.parse_args(sys.argv)

    jobs = find_jobs(options.xrslfile, '.')
    print('Found {} jobs'.format(len(jobs)))
    unfinished = [j for j in jobs if not is_finished(j)]
    print('Jobs without a result: {}'.format(len(unfinished)))

    if options.redo:
        print('All jobs will be executed')
    else:
        print('Only unfinished jobs will be executed')
        jobs = unfinished
        del unfinished

    if options.confirm:
        print('Program will continue when you press Enter. If something looks wrong kill it now.')
        input()

    n = len(jobs)
    root = os.getcwd()
    for i, job in enumerate(jobs):
        print("-----------------------------------------------------------")
        print("Working on job {} of {}".format(i+1, n))
        print("Folder: {}".format(job['folder']))
        os.chdir(job['folder'])
        exec_xrsl(job['file'], cleanup=options.cleanup)
        os.chdir(root)
