import torch as th
import numpy as np
from torch import nn
from torch.autograd import Function


class Embedding(nn.Module):
    def __init__(self, num_tokens, dim):
        super(Embedding, self).__init__()
        self.manifold = PoincareManifold()
        self.lt = nn.Embedding(num_tokens, dim)
        self.manifold.init_weights(self.lt.weight, 1e-4)

    def forward(self, sources, neighbours):
        sources = sources.repeat((neighbours.size(1), 1)).permute(1,0)
        # neighbours = neighbours.repeat((sources.size(0), 1)).permute(1,0)
        source_embs = self.lt(sources)
        neighbour_embs = self.lt(neighbours)
        return -self.manifold.distance(source_embs, neighbour_embs)

    def embedding(self):
        return list(self.lt.parameters())[0].data.cpu().numpy()

    def optim_params(self):
        return [{
            'params': self.lt.parameters(),
            'rgrad': self.manifold.rgrad,
            'expm': self.manifold.expm,
            'logm': self.manifold.logm,
            'ptransp': self.manifold.ptransp,
        }, ]

    def save_weights(self, path):
        embs = self.lt.weight
        np.save(path, embs.detach().cpu().numpy())



class PoincareManifold:

    def __init__(self, eps=1e-5):
        self.eps = eps
        self.max_norm = 1 - eps

    def norm_bound(self, u):
        d = u.size(-1)
        u.view(-1, d).renorm_(2, 0, self.max_norm)
        return u

    def normalize(self, u):
        return 1e-4 * nn.functional.normalize(u, dim=-1)

    def distance(self, u, v):
        return Distance.apply(u, v, self.eps)

    def pnorm(self, u, dim=-1):
        return (u * u).sum(dim=dim).sqrt()

    def rgrad(self, p, d_p):
        if d_p.is_sparse:
            p_sqnorm = th.sum(
                p[d_p._indices()[0].squeeze()] ** 2, dim=1,
                keepdim=True
            ).expand_as(d_p._values())
            n_vals = d_p._values() * ((1 - p_sqnorm) ** 2) / 4
            n_vals.renorm_(2, 0, 5)
            d_p = th.sparse.DoubleTensor(d_p._indices(), n_vals, d_p.size())
        else:
            p_sqnorm = th.sum(p ** 2, dim=-1, keepdim=True)
            d_p = d_p * ((1 - p_sqnorm) ** 2 / 4).expand_as(d_p)
        return d_p

    def expm(self, p, d_p, normalize=False, lr=None, out=None):
        if lr is not None:
            d_p.mul_(-lr)
        if out is None:
            out = p
        out.add_(d_p)
        if normalize:
            self.norm_bound(out)
        return out

    def logm(self, p, d_p, out=None):
        return p - d_p

    def ptransp(self, p, x, y, v):
        ix, v_ = v._indices().squeeze(), v._values()
        return p.index_copy_(0, ix, v_)

    def init_weights(self, w, scale=1e-4):
        w.data.uniform_(-scale, scale)

    @staticmethod
    def dim(dim):
        return dim


class Distance(Function):
    @staticmethod
    def grad(x, v, sqnormx, sqnormv, sqdist, eps):
        alpha = (1 - sqnormx)
        beta = (1 - sqnormv)
        z = 1 + 2 * sqdist / (alpha * beta)
        a = ((sqnormv - 2 * th.sum(x * v, dim=-1) + 1) / th.pow(alpha, 2))\
            .unsqueeze(-1).expand_as(x)
        a = a * x - v / alpha.unsqueeze(-1).expand_as(v)
        z = th.sqrt(th.pow(z, 2) - 1)
        z = th.clamp(z * beta, min=eps).unsqueeze(-1)
        return 4 * a / z.expand_as(x)

    @staticmethod
    def forward(ctx, u, v, eps):
        squnorm = th.clamp(th.sum(u * u, dim=-1), 0, 1 - eps)
        sqvnorm = th.clamp(th.sum(v * v, dim=-1), 0, 1 - eps)
        sqdist = th.sum(th.pow(u - v, 2), dim=-1)
        ctx.eps = eps
        ctx.save_for_backward(u, v, squnorm, sqvnorm, sqdist)
        x = sqdist / ((1 - squnorm) * (1 - sqvnorm)) * 2 + 1
        # arcosh
        z = th.sqrt(th.pow(x, 2) - 1)
        return th.log(x + z)

    @staticmethod
    def backward(ctx, g):
        u, v, squnorm, sqvnorm, sqdist = ctx.saved_tensors
        g = g.unsqueeze(-1)
        gu = Distance.grad(u, v, squnorm, sqvnorm, sqdist, ctx.eps)
        gv = Distance.grad(v, u, sqvnorm, squnorm, sqdist, ctx.eps)
        return g.expand_as(gu) * gu, g.expand_as(gv) * gv, None