import numpy as np
import os


class Node:
    def __init__(self, name):
        self.name = name
        self.parents = set()
        self.children = set()
        self.depth = None
        self.order = None
        self.description = None

    def __repr__(self):
        return "Node({})".format(self.name)

    def __hash__(self):
        return hash(self.name)

    def get_ancestors(self):
        nodes = set()
        new_nodes = {n for n in self.parents}
        while len(new_nodes) > 0:
            nodes |= new_nodes
            one_up = set()
            for n in new_nodes:
                one_up |= n.parents
            new_nodes = one_up

        return nodes

    def get_descendants(self):
        nodes = set()
        new_nodes = {n for n in self.children}
        while len(new_nodes) > 0:
            nodes |= new_nodes
            one_up = set()
            for n in new_nodes:
                one_up |= n.children
            new_nodes = one_up

        return nodes

    def collect_ancestors(self, collection, max_order):
        if self.order <= max_order:
            collection.add(self.order)
        else:
            for p in self.parents:
                p.collect_ancestors(collection, max_order)


class Hierarchy:
    def __init__(self, nodes=None):

        self.nodes_by_name = nodes
        self.components = None
        self.is_tree = None
        self.roots = None
        self.depth = None
        self.depth_discrepancy = None
        self.number_edges = None
        self.number_nodes = None
        self.idx = None
        self.node_order = []
        self.compute_stats()
        self.topological_sort()
        # self.print_stats()

    def assert_single_root(self):
        if len(self.roots) > 1:
            root = Node('I_am_root')
            self.nodes_by_name[root.name] = root
            for node in self.roots:
                root.children.add(node)
                node.parents.add(root)

            for n in self.nodes_by_name.values():
                n.depth = None
                n.order = None

            self.compute_stats()
            self.topological_sort()

    def prune(self, max_order):
        new_nodes = {}
        max_order = min(max_order, self.number_nodes-1)

        for i in range(max_order+1):
            node = self.node_order[i]
            new_node = Node(node.name)
            new_node.order = node.order
            new_node.depth = node.depth
            new_node.description = node.description
            for p in node.parents:
                new_node.parents.add(new_nodes[p.name])
                new_nodes[p.name].children.add(new_node)
            assert new_node.name not in new_nodes, "nodes with the same name"
            new_nodes[new_node.name] = new_node

        return Hierarchy(new_nodes)

    def sub_hierarchy(self, root_node, num_nodes):
        # seems problematic when DAGiness kicks in
        new_nodes = {}
        node_queue = [root_node]
        while len(node_queue) > 0 and len(new_nodes) < num_nodes:
            node = node_queue.pop(0)
            if node.name in new_nodes:
                continue
            new_node = Node(node.name)
            new_node.description = node.description
            for p in node.parents:
                if p.name in new_nodes:
                    new_node.parents.add(new_nodes[p.name])
                    new_nodes[p.name].children.add(new_node)
            for c in node.children:
                node_queue.append(c)
            assert new_node.name not in new_nodes, "nodes with the same name"
            new_nodes[new_node.name] = new_node
        return Hierarchy(new_nodes)

    @classmethod
    def parse_hier_file(cls, hier_path):
        relations = []
        with open(hier_path) as f:
            for line in f:
                words = line.strip().split()
                if words[0] == '@is_parent':
                    relations.append('{}/{}'.format(words[1], words[2]))

        return cls.parse_dag_hierarchy(','.join(relations))

    @classmethod
    def parse_obo_file(cls, obo_path):
        relations = []
        descriptions = {}
        with open(obo_path) as f:
            for line in f:
                if line.startswith('[Typedef]'):
                    break

                if line.startswith('id:'):
                    term = line.strip().split()[1].replace(':', '')

                elif line.startswith('is_a:'):
                    parent = line.strip().split()[1].replace(':', '')
                    relations.append('{}/{}'.format(parent, term))

                elif line.startswith('name:'):
                    descriptions[term] = line.strip()[6:]

                elif line.startswith('relationship: part_of'):
                    parent = line.strip().split()[2].replace(':', '')
                    relations.append('{}/{}'.format(parent, term))

        h = cls.parse_dag_hierarchy(','.join(relations))
        for node in h.node_order:
            node.description = descriptions[node.name]
        return h

    @classmethod
    def parse_arff_file(cls, arff_path, type):
        with open(arff_path) as f:
            for line in f:
                words = line.strip().split()
                if len(words) == 4 and words[0].lower() == '@attribute' and words[2].lower() == 'hierarchical':
                    arff_string = words[-1]
                    break

        if type == 'DAG':
            return cls.parse_dag_hierarchy(arff_string)
        elif type == 'Tree':
            return cls.parse_tree_hierarchy(arff_string)
        else:
            raise Exception('Unsuported hierarchy type!')

    @classmethod
    def parse_old_go_folder(cls, folder_path):
        files = os.listdir(folder_path)

        relations = []
        for f in files:
            parents = []
            previous_level = 0
            previous_term = None
            with open('{}/{}'.format(folder_path, f)) as g:
                for line in g:
                    stripped = line.strip()
                    if stripped[0] in ['%', '<']:
                        level = len(line) - len(stripped)
                        term = stripped.split(';')[1].split()[0].strip(',').replace(':', '')

                        if previous_term is not None and level > previous_level:
                            parents.append((previous_term, previous_level))
                        if level < previous_level:
                            while level <= parents[-1][1]:
                                parents.pop()
                        if previous_term is not None:
                            relations.append('{}/{}'.format(parents[-1][0], term))

                        previous_term = term
                        previous_level = level

        return cls.parse_dag_hierarchy(','.join(relations))

    @classmethod
    def parse_dag_hierarchy(cls, arff_string):
        pairs = arff_string.split(',')
        nodes = {}

        for p in pairs:
            n1, n2 = p.split('/')
            nodes.setdefault(n1, Node(n1))
            nodes.setdefault(n2, Node(n2))
            nodes[n1].children.add(nodes[n2])
            nodes[n2].parents.add(nodes[n1])

        return cls(nodes)

    @classmethod
    def parse_tree_hierarchy(cls, arff_string):
        nodes = {}

        def process_tree_label(self, l):
            nodes.setdefault(l, Node(l))
            if '/' in l:
                l1 = l[:l.rfind('/')]
                process_tree_label(l1)
                nodes[l].parents.add(nodes[l1])
                nodes[l1].children.add(nodes[l])

        labels = arff_string.split(',')
        for l in labels:
            process_tree_label(l)
        return cls(nodes)

    @classmethod
    def parse_edgelist(cls, edges_file, labels_file):
        nodes = {}
        with open(edges_file) as f:
            for line in f:
                n1, n2 = line.strip().split()
                node1 = nodes.setdefault(n1, Node(n1))
                node2 = nodes.setdefault(n2, Node(n2))
                node1.children.add(node2)
                node2.parents.add(node1)

        h = cls(nodes)

        with open(labels_file) as f:
            for line in f:
                order, name = line.strip().split()
                node = h.nodes_by_name[order]
                node.name = name
                node.order = int(order)
                h.node_order[node.order] = node

        return h

    def compute_stats(self):
        def depth_first_search(root_node):
            stack = [root_node]
            while stack:
                current = stack.pop()
                visited.add(current.name)
                for c in current.children | current.parents:
                    if c.name not in visited:
                        stack.append(c)

        self.is_tree = True
        self.roots = []
        for node in self.nodes_by_name.values():
            if len(node.parents) == 0:
                self.roots.append(node)
            elif len(node.parents) > 1:
                self.is_tree = False
        self.components = 0
        visited = set()
        for name, node in self.nodes_by_name.items():
            if name not in visited:
                self.components += 1
                depth_first_search(node)
        self.number_nodes = len(self.nodes_by_name)
        self.number_edges = sum(len(n.children) for n in self.nodes_by_name.values())
        self.compute_depths()

    def topological_sort(self):
        self.kahn_sort()

    def kahn_sort(self):
        self.node_order = [None]*self.number_nodes
        unvisited = [n for n in self.roots]
        order = 0

        while len(unvisited) > 0:
            n = unvisited.pop(0)
            self.node_order[order] = n
            n.order = order
            order += 1
            for m in n.children:
                m.parents.remove(n)
                if len(m.parents) == 0:
                    unvisited.append(m)

        for n in self.node_order:
            for c in n.children:
                c.parents.add(n)

    def dfs_sort(self):

        visited = {n: False for n in self.nodes_by_name}

        def visit(n):
            if not visited[n.name]:
                visited[n.name] = True
                for m in n.children:
                    visit(m)
                self.node_order[self.idx] = n
                n.order = self.idx
                self.idx -= 1

        self.node_order = [None] * self.number_nodes
        self.idx = len(self.nodes_by_name) - 1
        for n in self.nodes_by_name:
            visit(self.nodes_by_name[n])
        self.idx = None

    def compute_depths(self):
        def aggregate_depths(s):
            return np.mean([x.depth for x in s])

        def bfs(queue_of_nodes):
            while queue_of_nodes:
                new_queue = []
                for n in queue_of_nodes:
                    should_keep = True
                    if n.depth is None:
                        if not any([p.depth is None for p in n.parents]):
                            n.depth = 1 + aggregate_depths(n.parents)
                            should_keep = False
                    else:
                        should_keep = False
                    if should_keep:
                        new_queue.append(n)
                    else:
                        for c in n.children:
                            new_queue.append(c)
                queue_of_nodes = new_queue

        for r in self.roots:
            r.depth = 1.0
        bfs([c for root in self.roots for c in root.children])
        self.depth = -1
        self.depth_discrepancy = (-1, None)
        for node in self.nodes_by_name.values():
            self.depth = max(self.depth, node.depth)
            if node.parents:
                parents_depths = [x.depth for x in node.parents]
                d = max(parents_depths) - min(parents_depths)
                if d > self.depth_discrepancy[0]:
                    self.depth_discrepancy = (d, node)

    def print_stats(self):
        print("components", self.components)
        print("is_tree:", self.is_tree)
        print("roots:", self.roots)
        print("depth:", self.depth)
        print("depth discrepancy:", self.depth_discrepancy)
        print("nodes:", self.number_nodes)
        print("edges:", self.number_edges)

    def save_to_file(self, path):
        with open(path + '.labels', 'w') as f:
            for node in self.node_order:
                print(node.order, node.name, file=f)

        with open(path + '.edges', 'w') as f:
            for node in self.node_order:
                for child in node.children:
                    print(node.order, child.order, file=f)

    def set_to_vec(self, label_set):
        vec = np.zeros(self.number_nodes, dtype=np.float32)

        def mark_node(n):
            vec[n.order] = 1.
            for p in n.parents:
                mark_node(p)

        for l in label_set:
            mark_node(self.nodes_by_name[l])
        return vec

    def random_label_set(self, threshold=1e-4):
        rng = np.random.uniform(size=self.number_nodes)
        labels = {self.node_order[i].name for i, r in enumerate(rng) if r < threshold}
        return labels

    def neighbourhood_matrix(self, path_length=1):
        matrix = np.zeros((self.number_nodes, self.number_nodes), dtype=np.float32)

        def tag_parents(row, node, steps):
            matrix[row, node.order] = 1
            if steps > 0:
                for n in node.parents:
                    tag_parents(row, n, steps-1)

        def tag_children(row, node, steps):
            matrix[row, node.order] = 1
            if steps > 0:
                for n in node.children:
                    tag_children(row, n, steps-1)

        for node in self.nodes_by_name.values():
            tag_parents(node.order, node, path_length)
            tag_children(node.order, node, path_length)

        return matrix

    def sanity_ceck(self):
        if len(self.node_order) != self.number_nodes:
            print('node_order lenght {}, number of nodes {}'.format(len(self.node_order), self.number_nodes))

        if len(self.nodes_by_name) != self.number_nodes:
            print('nodes_by_name lenght {}, number of nodes {}'.format(len(self.nodes_by_name), self.number_nodes))

        for node in self.node_order:
            if node.name not in self.nodes_by_name:
                print(node.name, 'not found?')
            for c in node.children:
                if node not in c.parents:
                    print('child-parent not symetricall')
                if c.name not in self.nodes_by_name:
                    print('child not in the hierarchy')

            for p in node.parents:
                if p.name not in self.nodes_by_name:
                    print('parent not in the hierarchy')

        print('checks completed')